/***********************************************************************
Alcaze - A bitmap to spectrogram converter - Version 0.1
Copyright (C) 2007 Tito Dal Canton (tito at dalcanton.it)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>

int load_bitmap(const char *file_path, unsigned short *out_pixels, unsigned int *img_width, unsigned int *img_height);
unsigned char get8(FILE *in);
unsigned short get16(FILE *in);
unsigned long get32(FILE *in);

int main(int argc, char *argv[])
{
	unsigned int	img_width = 128,
					img_height = 512,
					samples_per_pixel = 100,
					base_freq = 2000,
					factor = 900,
					uNs = img_height * 44100 / samples_per_pixel,
					Ns = uNs / factor,
					i, j;
	short			**samples = NULL,
					*upsamples = NULL;
	unsigned short	*img = NULL;
	float			*sincbuffer = NULL;

	if (argc < 2) {
		fprintf(stderr, "Alcaze - A bitmap to spectrogram converter - Version 0.1\n"
						"Copyright (C) 2007 Tito Dal Canton (tito at dalcanton.it).\n\n"
						"This program is free software; you can redistribute it\n"
						"and/or modify it under the terms of the GNU General Public\n"
						"License. For details, see the file COPYING.\n\n"
						"Usage:\n\n"
						"    alcaze <bitmap file path>\n\n");
		return 0;
	}

	/* allocation of the bitmap buffer */
	img = (unsigned short *)malloc(img_width * img_height * sizeof(unsigned short));
	if (img == NULL)
		return -1;

	/* load the bitmap */
	fprintf(stderr, "Loading bitmap\n");
	if (load_bitmap(argv[1], img, &img_width, &img_height))
		return -1;

	/* allocate signal buffers */
	samples = (short **)malloc(img_width * sizeof(short *));
	if (samples == NULL)
		return -1;
	for (j = 0; j < img_width; j++) {
		samples[j] = (short *)malloc(Ns * sizeof(short));
		if (samples[j] == NULL)
			return -1;
	}
	upsamples = (short *)malloc(uNs * sizeof(short));
	if (upsamples == NULL)
		return -1;
	sincbuffer = (float *)malloc(Ns * sizeof(float));
	if (sincbuffer == NULL)
		return -1;

	/* narrowband noise signal generation */
	fprintf(stderr, "Generating narrowband noise\n");
	for (j = 0; j < img_width; j++) {
		for (i = 0; i < Ns; i++)
			samples[j][i] = (short)((65535.0 * rand()) / RAND_MAX - 32767.0);
	}

	/* 44100 Hz resampling, upconversion & modulation */
	fprintf(stderr, "Computing output signal (this will take time)\n");
	for (i = 0; i < uNs; i++) {
		unsigned int	x, y = i * samples_per_pixel / 44100;
		float			output = 0;
		short			output_int;

		/* precalculate the coefficients to resample narrowband signals */
		for (j = 0; j < Ns; j++) {
			int delta = i - j * factor;

			if (abs(delta) < 200 * factor) {
				/* sampling theorem interpolation formula */
				if (delta == 0)
					sincbuffer[j] = 1.0;
				else
					sincbuffer[j] = sin(M_PI * delta / factor) / (M_PI * delta / factor);
			} else {
				sincbuffer[j] = 0;
			}
		}
		/* upconvert and modulate */
		for (x = 0; x < img_width; x++) {
			unsigned int	j;
			float			s = 0;
			int				freq = base_freq + x * 44100 / factor;

			if (img[y*img_width + x] == 0)
				continue;

			/* resample the narrowband noise at 44100 Hz using the precalculated coefficients */
			for (j = 0; j < Ns; j++)
				s += samples[x][j] * sincbuffer[j];

			/* upconvert the resampled narrowband noise and time-modulate
			   it with the corresponding column of pixel values */
			/* FIXME large jumps between adjacent pixels cause wideband spikes at pixel
			   boundaries which look really bad on the spectrogram; smooth them */
			output += img[y*img_width + x] * s * cos(2.0 * M_PI * freq * i / 44100.0);
		}

		/* FIXME store the output samples, then go back and rescale
		   them all so that we output a normalized signal */
		output *= 4.0 / (img_width * 65535);
		output_int = (short)output;
		fwrite(&output_int, sizeof(short), 1, stdout);
	}

	free(samples);
	return 0;
}

int load_bitmap(const char *file_path, unsigned short *out_pixels, unsigned int *img_width, unsigned int *img_height)
{
	FILE			*in = NULL;
	unsigned int	i, x, y;
	unsigned short	bmp_magic, bpp, num_of_planes;
	unsigned long	bmp_size, header_size, compression, comp_size, color_count,
					bmp_width, bmp_height, img_offset;
	unsigned short	*graytable = NULL;

	/* open and check input file */
	in = fopen(file_path, "rb");
	if (in == NULL) {
		fprintf(stderr, "Can't open \"%s\" for reading: %s\n", file_path, strerror(errno));
		return -1;
	}
	/* bitmap header */
	bmp_magic = get16(in);
	if (bmp_magic != 0x4d42) {
		fprintf(stderr, "Sorry, \"%s\" is not a BMP bitmap\n", file_path);
		return -1;
	}
	bmp_size = get32(in);
	get16(in);
	get16(in);
	img_offset = get32(in);
	/* bitmap info header */
	header_size = get32(in);
	bmp_width = get32(in);
	bmp_height = get32(in);
	num_of_planes = get16(in);
	bpp = get16(in);
	compression = get32(in);
	comp_size = get32(in);
	get32(in);
	get32(in);
	color_count = get32(in);
	get32(in);
	if (header_size != 40 || num_of_planes != 1) {
		fprintf(stderr, "Sorry, \"%s\" is not a BMP bitmap\n", file_path);
		return -1;
	}
	if (bpp != 8 || color_count > 256) {
		fprintf(stderr, "Sorry, the bitmap must be 8-bit color\n");
		return -1;
	}
	fprintf(stderr, "Bitmap dimensions: %lux%lux%u\n", bmp_width, bmp_height, bpp);
	/* read palette and build the gray level table */
	/* TODO log intensity scale correction (e.g. for baudline) */
	graytable = (unsigned short *)malloc(color_count * sizeof(unsigned short));
	if (graytable == NULL)
		return -1;
	for (i = 0; i < color_count; i++) {
		unsigned int	r, g, b;
		
		b = get8(in);
		g = get8(in);
		r = get8(in);
		get8(in);
		graytable[i] = sqrt(r*r + b*b + g*g) * 20000 / 442;
	}
	/* load pixels */
	fseek(in, img_offset, SEEK_SET);
	for (y = 0; y < bmp_height; y++) {
		if (y >= *img_height)
			break;
		for (x = 0; x < bmp_width; x++) {
			unsigned char	value = get8(in);

			if (x < *img_width)
				out_pixels[(*img_height - 1 - y) * (*img_width) + x] = graytable[value];
		}
	}
	free(graytable);
	fclose(in);

	return 0;
}

/* little-endian binary file access */

unsigned char get8(FILE *in)
{
	return fgetc(in);
}

unsigned short get16(FILE *in)
{
	unsigned short  lo = fgetc(in) & 0xff,
					hi = fgetc(in) & 0xff;
	
	return (hi << 8) | lo;
}

unsigned long get32(FILE *in)
{
	unsigned short  a = fgetc(in) & 0xff,
					b = fgetc(in) & 0xff,
					c = fgetc(in) & 0xff,
					d = fgetc(in) & 0xff;
	
	return (d << 24) | (c << 16) | (b << 8) | a;
}

